from __future__ import absolute_import

import os
from lxml import etree
import unicodedata

PUNCS_WE_DONT_LIKE = "[],.()<>'/?;:\""

def kpick(args, options, default=None):
    for opt in options:
        if args.get(opt, None):
            return args[opt]
    return default

def asciify(inp):
    '''nuke all the unicode from orbit. it's the only way to be sure.'''
    if inp:
        try:
            return inp.encode('ascii', 'ignore')
        except UnicodeDecodeError:
            return unicodedata.normalize('NFKD', inp.decode('utf-8')).encode('ascii', 'ignore')
    else:
        return ''

def parameterize(inp, sep='+'):
    '''make strings suitable for submission to GET-based query service. strips
        out these characters: %s''' % PUNCS_WE_DONT_LIKE
    return asciify(inp).replace(' ', sep).translate(None, PUNCS_WE_DONT_LIKE)

def deparameterize(inp, sep='+'):
    '''somewhat-undo parameterization in string. replace separators (sep) with spaces.'''
    return inp.replace(sep, ' ')

def remove_html_markup(s):
    '''remove html and xml tags from text. preserves HTML entities like &amp;'''
    tag = False
    quote = False
    out = ""

    for c in s:
            if c == '<' and not quote:
                tag = True
            elif c == '>' and not quote:
                tag = False
            elif (c == '"' or c == "'") and tag:
                quote = not quote
            elif not tag:
                out = out + c
    return out





import unittest

from metapub import FindIt
#from metapub.pubmedcentral import *

wiley_pmids = []
jstage_pmids = []
pmc_pmids = []
sciencedirect_pmids = []
bmj_pmids = []
oxford_pmids = []
lancet_pmids = []


class TestFindIt(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

 



from __future__ import absolute_import

from lxml import etree

class MetaPubObject(object):

    def __init__(self, xmlstr, root=None, *args, **kwargs):
        if not xmlstr:
            if xmlstr == '':
                xmlstr = 'empty'
            raise MetaPubError('Cannot build MetaPubObject; xml string was %s' % xmlstr)
        self.xmlstr = xmlstr
        self.content = self._parse_xml(xmlstr, root)

    def _parse_xml(self, xmlstr, root=None):
        dom = etree.fromstring(xmlstr)
        if root:
            return dom.find(root)
        else:
            return dom

    def _get(self, tag):
        n = self.content.find(tag)
        if n is not None:
            return n.text
        return None

# singleton class used by the fetchers.
class Borg(object):
    _shared_state = {}
    def __init__(self):
        self.__dict__ = self._shared_state





from __future__ import absolute_import

"""metapub.PubMedFetcher -- tools to deal with NCBI's E-utilities interface to PubMed"""

import xml.etree.ElementTree as ET
from eutils.exceptions import EutilsBadRequestError
import requests

from .pubmedarticle import PubMedArticle
from .pubmedcentral import get_pmid_for_otherid
from .utils import kpick, parameterize
from .text_mining import re_pmid
from .exceptions import *
from .base import Borg
from .config import DEFAULT_EMAIL

class PubMedFetcher(Borg):
    '''PubMedFetcher (a Borg singleton object)

    An interaction layer for querying via specified method to return PubMedArticle objects.
    
    Currently available methods: eutils

    Basic Usage:

        fetch = PubMedFetcher()
    
    To specify a service method (more coming soon):

        fetch = PubMedFetcher('eutils')

    To return an article by querying the service with a known PMID:

        paper = fetch.article_by_pmid('123456')

    Similar methods exist for returning papers by DOI and PM Central id:

        paper = fetch.article_by_doi('10.1038/ng.379')
        paper = fetch.article_by_pmcid('PMC3458974')

    Finally, you can search for PMIDs via citation details by using the pmids_for_citation
    method, for which you only need 3 out of 5 details as a list of one or more PMIDs:

        pmids = fetch.pmids_for_citation(journal='Science', year='2008', volume='4', 
                first_page='7', author_name='Grant')
    '''

    def __init__(self, method='eutils', email=DEFAULT_EMAIL):
        Borg.__init__(self)
        self.method = method

        if method=='eutils':
            import eutils.client as ec
            self.qs = ec.QueryService(tool='metapub', email=email)
            self.article_by_pmid = self._eutils_article_by_pmid
            self.article_by_pmcid = self._eutils_article_by_pmcid
            self.article_by_doi = self._eutils_article_by_doi
            self.pmids_for_query = self._eutils_pmids_for_query
        else:
            raise NotImplementedError('coming soon: fetch from local pubmed via medgen-mysql or filesystem cache.')

    def _eutils_article_by_pmid(self, pmid):
        pmid = str(pmid)
        try:
            result = self.qs.efetch(args={'db': 'pubmed', 'id': pmid})
        except EutilsBadRequestError:
            raise MetaPubError('Invalid ID "%s" (rejected by Eutils); please check the number and try again.' % pmid)

        if result==None:
            return None
        if result.find('ERROR') > -1:
            raise MetaPubError('PMID %s returned ERROR; cannot construct PubMedArticle' % pmid)

        try:
            return PubMedArticle(result)
        except AttributeError:
            # in this case, eutils let us fetch a good-looking pmid, but it did not parse as an article.
            raise InvalidPMID('Pubmed ID "%s" not found.' % pmid)

    def _eutils_article_by_pmcid(self, pmcid):
        # if user submitted a bare number, prepend "PMC" to make sure it is submitted correctly 
        # the conversion API at pubmedcentral.
        pmcid = str(pmcid)
        if re_pmid.findall(pmcid)[0] == pmcid:
            pmcid = 'PMC'+pmcid

        pmid = get_pmid_for_otherid(pmcid)
        if pmid is None:
            raise MetaPubError('No PMID available for PubMedCentral id %s' % pmcid)
        return self._eutils_article_by_pmid(pmid)
    
    def _eutils_article_by_doi(self, doi):
        pmid = get_pmid_for_otherid(doi)
        if pmid is None:
            raise MetaPubError('No PMID available for doi %s' % doi)
        return self._eutils_article_by_pmid(pmid)

    def _eutils_pmids_for_query(self, query='', **kwargs):
        '''returns list of pmids for given freeform query string plus 
            keyword arguments.'''
        q = {}
        q['AID'] = kpick(kwargs, options=['AID', 'doi']) 
        q['PMC'] = kpick(kwargs, options=['pmc', 'pmcid'])
        q['TA'] = kpick(kwargs, options=['TA', 'jtitle', 'journal']) 
        q['DP'] = kpick(kwargs, options=['DP', 'year', 'pdat'])
        q['1AU'] = kpick(kwargs, options=['1AU', 'aulast', 'author1_lastfm', 'author1_last_fm'])
        q['OT'] = kwargs.get('OT', None)
        q['PT'] = kpick(kwargs, options=['PT', 'pubmed_type'])
        q['MH'] = kpick(kwargs, options=['MH', 'mesh', 'MeSH Terms'])
        q['FAU'] = kpick(kwargs, options=['FAU', 'first_author', 'author1'])
        q['IP'] = kpick(kwargs, options=['IP', 'issue'])
        q['TI'] = kpick(kwargs, options=['TI', 'title', 'atitle', 'article_title'])
        q['TW'] = kpick(kwargs, options=['TW', 'text'])
        q['LA'] = kpick(kwargs, options=['LA', 'language'])
        q['VI'] = kpick(kwargs, options=['VI', 'volume', 'vol'])
        q['PUBN'] = kpick(kwargs, options=['PUBN', 'publisher'])
        q['book'] = kwargs.get('book', None)
        q['ISBN'] = kwargs.get('ISBN', None)
        q['LASTAU'] = kwargs.get('LASTAU', None)

        #MeSH Date [MHDA]
        #MeSH Major Topic [MAJR]
        #MeSH Subheadings [SH]
        #MeSH Terms [MH]        
        #q['MHDA'] = kwargs.get('MHDA', None)

        q['PMID'] = kpick(kwargs, options=['pmid', 'uid', 'pubmed_id'])
        
        if query != '':
            query = query + ' '
        for feature in q.keys():
            if q[feature] != None:
                query +='%s[%s] ' % (q[feature], feature)
        
        # option to query pubmed central only:
        # pubmed pmc[sb]
        if kwargs.get('pmc_only', False):
            query += 'pubmed pmc[sb]'

        results = self.qs.esearch({'db': 'pubmed', 'term': query})
        return results


    def pmids_for_citation(self, **kwargs):
        '''returns list of pmids for given citation. requires at least 3/5 of these keyword arguments:
            jtitle or journal (journal title)
            year or date
            volume
            spage or first_page (starting page / first page)
            aulast (first author's last name) or author1_first_lastfm (as produced by PubMedArticle class)

        (Note that these arguments were made to match the tokens that arise from CrossRef's result['slugs'].)
        '''
        # output format in return:
        # journal_title|year|volume|first_page|author_name|your_key|
        base_uri = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/ecitmatch.cgi?db=pubmed&retmode=xml&bdata={journal_title}|{year}|{volume}|{first_page}|{author_name}|metapub|'

        journal_title = kpick(kwargs, options=['jtitle', 'journal', 'journal_title'], default='')
        author_name = _reduce_author_string(kpick(kwargs, 
                        options=['aulast', 'author1_last_fm', 'author', 'authors'], default=''))
        first_page = kpick(kwargs, options=['spage', 'first_page'], default='')
        year = kpick(kwargs, options=['year', 'date', 'pdat'], default='')
        volume = kpick(kwargs, options=['volume'], default='')

        inp_dict = { 'journal_title': parameterize(journal_title, '+'), 
                     'year': str(year),
                     'volume': str(volume),
                     'first_page': str(first_page),
                     'author_name': parameterize(author_name, '+'),
                   }

        req = base_uri.format(**inp_dict)
        content = requests.get(req).text
        pmids = []
        for item in content.split('\n'):
            if item.strip():
                pmid = item.split('|')[-1]
                pmids.append(pmid.strip())
        return pmids
                

def _reduce_author_string(author_string):
    # try splitting by commas
    authors = author_string.split(',')
    if len(authors)<2:
        # try splitting by semicolons
        authors = author_string.split(';')

    author1 = authors[0]
    # presume last name is at the end of the string
    return author1.split(' ')[-1]


""" 
Search Field Descriptions and Tags

from http://www.ncbi.nlm.nih.gov/books/NBK3827/

Affiliation [AD]
Article Identifier [AID]
All Fields [ALL]
Author [AU]
Author Identifier [AUID]
Book [book]
Comment Corrections
Corporate Author [CN]
Create Date [CRDT]
Completion Date [DCOM]
EC/RN Number [RN]
Editor [ED]
Entrez Date [EDAT]
Filter [FILTER]
First Author Name [1AU]
Full Author Name [FAU]
Full Investigator Name [FIR]
Grant Number [GR]   Investigator [IR]
ISBN [ISBN]
Issue [IP]
Journal [TA]
Language [LA]
Last Author [LASTAU]
Location ID [LID]
MeSH Date [MHDA]
MeSH Major Topic [MAJR]
MeSH Subheadings [SH]
MeSH Terms [MH]
Modification Date [LR]
NLM Unique ID [JID]
Other Term [OT]
Owner
Pagination [PG]
Personal Name as Subject [PS]   Pharmacological Action [PA]
Place of Publication [PL]
PMID [PMID]
Publisher [PUBN]
Publication Date [DP]
Publication Type [PT]
Secondary Source ID [SI]
Subset [SB]
Supplementary Concept[NM]
Text Words [TW]
Title [TI]
Title/Abstract [TIAB]
Transliterated Title [TT]
UID [PMID]
Version
Volume [VI]
"""




import sys

from metapub import PubMedFetcher

# examples of different formats:
# 18612690: PubMedArticle with multiple AbstractText sections
# 1234567:  PubMedArticle with no abstract whatsoever
# 20301546: PubMedBookArticle from GeneReviews

####
import logging
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("eutils").setLevel(logging.WARNING)
####

try:
    pmid = sys.argv[1]
except IndexError:
    print "Supply a pubmed ID as the argument to this script."
    print ""
    print "Example: python demo_pubmed.py 123456"
    sys.exit()

fetch = PubMedFetcher()

article = fetch.article_by_pmid(pmid)
print article.pmid, article.title
#print 'pages: '+article.pages
print 'authors: %s' % ','.join(article.authors)
print 'abstract: %s' % article.abstract
if article.pii:
    print 'pii: '+article.pii
if article.doi:
    print 'doi: '+article.doi

#print article.xmlstr




from metapub import PubMedFetcher
fetch = PubMedFetcher()
params = { 'jtitle': 'American Journal of Medical Genetics', 
                    'year': 1996, 
                    'volume': 61, 
                    'spage': 10, 
                    'authors': 'Katherine M. Hegmann; Aimee S. Spikes; Avi Orr-Urtreger; Lisa G. Shaffer' }

stuff = fetch.pmids_for_citation(**params)
print params
print stuff

params = { 'jtitle':'Journal of Neural Transmission', 
                    'year':2014, 
                    'volume':121, 
                    'first_page':1077, 
                    # author_name='Freitag'
         } 

stuff = fetch.pmids_for_citation(**params)

print params
print stuff




from __future__ import absolute_import

"""metapub.MedGenFetcher -- tools to deal with NCBI's E-utilities interface to the MedGen db"""

from lxml import etree

from .exceptions import MetaPubError
from .medgenconcept import MedGenConcept
from .base import Borg
from .config import DEFAULT_EMAIL

class MedGenFetcher(Borg):
    '''MedGenFetcher (a Borg singleton object)

    An interaction layer for querying to return MedGenConcept objects.
    
    Currently available methods: eutils

    Basic Usage:

        fetch = MedGenFetcher()
    
    To specify a service method (more coming soon):

        fetch = MedGenFetcher('eutils')

    To return a MedGenConcept from a known UID:

        concept = fetch.concept_by_uid(known_UID)

    To return a list of UIDs relevant to a given term known in medgen:

        uids = fetch.uids_by_term(some_term)

    To get a medgen UID given a known Concept ID (cui):

        uid = fetch.uid_for_cui(known_cui)
    '''

    def __init__(self, method='eutils', email=DEFAULT_EMAIL):
        Borg.__init__(self)
        self.method = method

        if method=='eutils':
            import eutils.client as ec
            self.qs = ec.QueryService(email=email)
            self.uids_by_term = self._eutils_uids_by_term
            self.concept_by_uid = self._eutils_concept_by_uid
            self.uid_for_cui = self._eutils_uid_for_cui
        else:
            raise NotImplementedError('coming soon: fetch from local medgen via medgen-mysql.')

    def _eutils_uids_by_term(self, term):
        '''wraps results of an medgen efetch term lookup, returning IDs of related MedGenConcepts.
        
        :return: list of medgen uids (strings)
        '''
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=medgen&term=OCRL
        result = self.qs.esearch( { 'db': 'medgen', 'term': term } )
        dom = etree.fromstring(result)
        uids = []
        idlist = dom.find('IdList')
        for item in idlist.findall('Id'):
            uids.append(item.text.strip())
        return uids
    
    def _eutils_concept_by_uid(self, uid):
        '''wraps results of an esummary fcgi call to medgen when ID is known.
        
        :return: MedGenConcept object
        '''
        # http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=medgen&id=336867
        uid = str(uid)
        result = self.qs.esummary( { 'db': 'medgen', 'id': uid } )
        return MedGenConcept(result)

    def _eutils_uid_for_cui(self, cui):
        '''given a ConceptID (cui), return a medgen ID.'''
        #http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=medgen&term=C0000039
        if not cui.startswith('C'):
            raise MetaPubError('Invalid CUID: must start with C (e.g. C0000039)')

        result = self.qs.esearch( { 'db': 'medgen', 'term': cui } )
        root = etree.fromstring(result).getroottree()
        try:
            uid = root.find('IdList').find('Id').text.strip()
        except AttributeError:
            raise MetaPubError('Invalid CUID: did not return MedGen id.')
        return uid




import sys

from tabulate import tabulate
from metapub import MedGenFetcher

try:
    input_gene = sys.argv[1]
except IndexError:
    print 'Supply a Hugo gene name to this script as its argument.'
    sys.exit()

####
import logging
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("eutils").setLevel(logging.INFO)
####

fetch = MedGenFetcher()
uids = fetch.uids_by_term(input_gene+'[gene]')
print uids

# TODO: Term Hierarchy Children (only 1 tier below), Term Hierarchy Parents (only 1 tier above)

headers = ['CUI', 'Hugo', 'Title', 'Semantic Type', 'MedGenUID', 
           'OMIM ID', 'Modes of Inheritance', 'Assoc Genes', ]

table = []

def _join_or_NA(some_list, select=None, joiner=','):
    'returns a joined string or NA if empty'
    if some_list and select:
        return joiner.join(item[select] for item in some_list)
    elif some_list:
        return joiner.join([item for item in some_list])
    else:
        return 'NA'

for this_id in uids:
    concept = fetch.concept_by_uid(this_id)
    #print concept.to_dict()
    assert concept.medgen_uid == this_id
    line = [concept.CUI, input_gene, concept.title, concept.semantic_type, concept.medgen_uid]

    line.append(_join_or_NA(concept.OMIM))
    line.append(_join_or_NA(concept.modes_of_inheritance, 'name'))
    line.append(_join_or_NA(concept.associated_genes, 'hgnc')) 

    table.append(line)

print tabulate(table, headers, tablefmt="simple")

